import logging
import os
import re

import pandas as pd
import rpy2.robjects as ro
import rpy2.robjects as robjects
from rpy2.rinterface_lib.sexp import NULLType
from rpy2.robjects import pandas2ri
from rpy2.robjects.conversion import localconverter
from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import FactorVector

from . import checkatlas, folders
from .metrics import metrics

"""
Module for management of Atlas n Seurat format

- Adata and summary tables are created from seuratà
- QC are created using Seurat
- UMAP, t-SNE are created using Seurat
- For creating metrics tables, obs tables are
  first converted from R to python and then
  scikitlearn is used.
"""

logger = logging.getLogger("checkatlas")

SCANPY_TO_SEURAT_OBS = {
    "total_counts": "nCount_RNA",
    "n_genes_by_counts": "nFeature_RNA",
    "pct_counts_mt": "percent.mito",
    "pct_counts_ribo": "percent.ribo",
    "violin_plot": "violin_plot",
}

SEURAT_TO_SCANPY_OBS = {
    "nCount_RNA": "total_counts",
    "nFeature_RNA": "n_genes_by_counts",
    "percent.mito": "pct_counts_mt",
    "percent.ribo": "pct_counts_ribo",
}


def check_seurat_install():
    """
    Check if Seurat is installed, run installation if not
    :return:
    """
    r_script = """install.packages(setdiff(c(\'Seurat\',\'SeuratObject\'),
                rownames(installed.packages())))"""
    robjects.r(r_script)


def read_atlas(atlas_path, atlas_info):
    """
    Read Seurat object in python using rpy2
    :param atlas_path:
    :param atlas_info:
    :return:
    """
    importr("Seurat")
    importr("SeuratObject")
    logger.info(f"Load {atlas_info[0]} in {atlas_info[-1]}")
    rcode = f'readRDS("{atlas_path}")'
    seurat = robjects.r(rcode)
    rclass = robjects.r["class"]
    if rclass(seurat)[0] == "Seurat":
        importr("Seurat")
        return seurat
    else:
        logger.info(f"{atlas_info[0]} is not a Seurat object")
        return None


def get_viable_obs_qc(seurat, args):
    """
    Search in obs_keys a match to OBS_QC values
    Extract sorted obs_keys in same order then OBS_QC
    :param adata:
    :return:
    """
    r_obs = robjects.r(
        "obs <- function(seurat){ return(colnames(seurat@meta.data))}"
    )
    obs_keys = list()
    for obs_qc in args.qc_display:
        obs_qc = SCANPY_TO_SEURAT_OBS[obs_qc]
        if obs_qc in r_obs(seurat):
            obs_keys.append(obs_qc)
    return obs_keys


def get_viable_obs_annot(seurat, args):
    """
    Search in obs_keys a match to OBS_CLUSTERS values
    ! Remove obs_key with only one category !
    Extract sorted obs_keys in same order then OBS_CLUSTERS
    :param adata:
    :return:
    """
    obs_keys = list()
    r_obs = robjects.r(
        "obs <- function(seurat){ return(colnames(seurat@meta.data))}"
    )
    obs_key_seurat = r_obs(seurat)
    r_annot = robjects.r(
        "type <- function(seurat, obs_key){ "
        "return(seurat[[obs_key]][[obs_key]])}"
    )
    # Get keys from OBS_CLUSTERS
    for obs_key in obs_key_seurat:
        for obs_key_celltype in args.obs_cluster:
            if obs_key_celltype in obs_key:
                if isinstance(r_annot(seurat, obs_key), FactorVector):
                    obs_keys.append(obs_key)
    # Remove keys with only one category
    obs_keys_final = list()
    for obs_key in obs_keys:
        annotations = r_annot(seurat, obs_key)
        if len(annotations.levels) != 1:
            logger.debug(
                f"Add obs_key {obs_key} with cat {annotations.levels}"
            )
            obs_keys_final.append(obs_key)
    return sorted(obs_keys_final)


def get_viable_obsm(seurat, args):
    """
    Search viable obsm for dimensionality reduction metric
    calc.
    ! No filter on osbm is appled for now !
    :param seurat:
    :param args:
    :return:
    """
    obsm_keys = list()
    # for obsm_key in adata.obsm_keys():
    #   if obsm_key in args.obsm_dimred:
    r_obsm = robjects.r(
        "f<-function(seurat){return(names(seurat@reductions))}"
    )
    obsm_keys = r_obsm(seurat)
    logger.debug(f"Add obsm {obsm_keys}")
    return obsm_keys


def create_summary_table(seurat, atlas_path, atlas_info, args) -> None:
    """
    Create a table with all interesting variables
    :param seurat:
    :param atlas_name:
    :param csv_path:
    :return:
    """
    atlas_name = atlas_info[0]
    logger.debug(f"Create Summary table for {atlas_name}")
    atlas_file_type = atlas_info[1]
    atlas_extension = atlas_info[2]
    csv_path = os.path.join(
        folders.get_folder(args.path, folders.SUMMARY),
        atlas_name + checkatlas.SUMMARY_EXTENSION,
    )
    # Create summary table
    header = [
        "AtlasFileType",
        "NbCells",
        "NbGenes",
        "AnnData.raw",
        "AnnData.X",
        "File_extension",
        "File_path",
    ]
    r_nrow = robjects.r["nrow"]
    r_ncol = robjects.r["ncol"]
    ncells = r_ncol(seurat)[0]
    ngenes = r_nrow(seurat)[0]
    x_raw = False
    x_norm = True
    df_summary = pd.DataFrame(index=[atlas_name], columns=header)
    df_summary["AtlasFileType"][atlas_name] = atlas_file_type
    df_summary["NbCells"][atlas_name] = ncells
    df_summary["NbGenes"][atlas_name] = ngenes
    df_summary["AnnData.raw"][atlas_name] = x_raw
    df_summary["AnnData.X"][atlas_name] = x_norm
    df_summary["File_extension"][atlas_name] = atlas_extension
    df_summary["File_path"][atlas_name] = atlas_path.replace(args.path, "")
    df_summary.to_csv(csv_path, index=False, sep="\t")


def create_anndata_table(seurat, atlas_path, atlas_info, args) -> None:
    """
    Create a table with all AnnData-like arguments in Seurat object
    :param seurat:
    :param atlas_name:
    :param atlas_path:
    :return:
    """
    atlas_name = atlas_info[0]
    logger.debug(f"Create Adata table for {atlas_name}")
    csv_path = os.path.join(
        folders.get_folder(args.path, folders.ANNDATA),
        atlas_name + checkatlas.ADATA_EXTENSION,
    )
    # Create AnnData table
    header = ["obs", "obsm", "var", "varm", "uns"]
    df_summary = pd.DataFrame(index=[atlas_name], columns=header)

    # Create r_functions
    r_obs = robjects.r(
        "obs <- function(seurat){ return(colnames(seurat@meta.data))}"
    )
    r_obsm = robjects.r(
        "f<-function(seurat){return(names(seurat@reductions))}"
    )
    r_uns = robjects.r(
        "uns <- function(seurat){ return(colnames(seurat@misc))}"
    )
    obs_list = r_obs(seurat)
    obsm_list = r_obsm(seurat)
    var_list = [""]
    varm_list = [""]
    uns_list = [""]
    if not isinstance(r_uns(seurat), NULLType):
        uns_list = r_uns(seurat)

    df_summary["obs"][atlas_name] = (
        "<code>" + "</code><br><code>".join(obs_list) + "</code>"
    )
    df_summary["obsm"][atlas_name] = (
        "<code>" + "</code><br><code>".join(obsm_list) + "</code>"
    )
    df_summary["var"][atlas_name] = (
        "<code>" + "</code><br><code>".join(var_list) + "</code>"
    )
    df_summary["varm"][atlas_name] = (
        "<code>" + "</code><br><code>".join(varm_list) + "</code>"
    )
    df_summary["uns"][atlas_name] = (
        "<code>" + "</code><br><code>".join(uns_list) + "</code>"
    )
    df_summary.to_csv(csv_path, index=False, quoting=False, sep="\t")


def create_qc_tables(seurat, atlas_path, atlas_info, args) -> None:
    """
    Display the atlas QC of seurat
    Search for the metadata variable which correspond
    to the total_RNA, total_UMI, MT_ratio, RT_ratio
    :param path:
    :param adata:
    :param atlas_name:
    :param atlas_path:
    :return:
    """
    atlas_name = atlas_info[0]
    qc_path = os.path.join(
        folders.get_folder(args.path, folders.QC),
        atlas_name + checkatlas.QC_EXTENSION,
    )
    logger.debug(f"Create QC tables for {atlas_name}")
    obs_keys = get_viable_obs_qc(seurat, args)
    r_meta = robjects.r("obs <- function(seurat){ return(seurat@meta.data)}")
    r_metadata = r_meta(seurat)
    with localconverter(ro.default_converter + pandas2ri.converter):
        df_metadata = ro.conversion.rpy2py(r_metadata)
        df_annot = df_metadata[obs_keys]
        # rename columns with scanpy names
        new_columns = list()
        for column in df_annot.columns:
            new_columns.append(SEURAT_TO_SCANPY_OBS[column])
        df_annot.columns = new_columns
        df_annot.to_csv(qc_path, index=False, quoting=False, sep="\t")


def create_qc_plots(seurat, atlas_path, atlas_info, args) -> None:
    """
    Display the atlas QC
    Search for the OBS variable which correspond to the toal_RNA, total_UMI,
     MT_ratio, RT_ratio
    :param path:
    :param adata:
    :param atlas_name:
    :param atlas_path:
    :return:
    """
    atlas_name = atlas_info[0]
    qc_path = os.path.join(
        folders.get_folder(args.path, folders.QC_FIG),
        atlas_name + checkatlas.QC_FIG_EXTENSION,
    )
    logger.debug(f"Create QC violin plot for {atlas_name}")
    importr("ggplot2")
    r_cmd = (
        "vln_plot <- function(seurat, obs, qc_path){"
        "vln <- VlnPlot(seurat, features = obs, ncol = length(obs));"
        "ggsave(qc_path, vln, width = 10, "
        "height = 4, dpi = 150)}"
    )
    r_violin = robjects.r(r_cmd)
    obs_keys = list(SEURAT_TO_SCANPY_OBS.keys())
    r_obs = robjects.StrVector(obs_keys)
    r_violin(seurat, r_obs, qc_path)


def create_umap_fig(seurat, atlas_path, atlas_info, args) -> None:
    """
    Display the UMAP of celltypes
    Search for the OBS variable which correspond to the celltype annotation
    :param path:
    :param adata:
    :param atlas_name:
    :param atlas_path:
    :return:
    """
    atlas_name = atlas_info[0]
    # Search if tsne reduction exists
    r = re.compile(".*umap*.")
    r_names = robjects.r["names"]
    obsm_list = r_names(seurat)
    importr("ggplot2")
    if len(list(filter(r.match, obsm_list))) > 0:
        logger.debug(f"Create UMAP figure for {atlas_name}")
        # Setting up figures directory
        umap_path = os.path.join(
            folders.get_folder(args.path, folders.UMAP),
            atlas_name + checkatlas.UMAP_EXTENSION,
        )
        # Exporting umap
        obs_keys = get_viable_obs_annot(seurat, args)
        r_cmd = (
            "umap <- function(seurat, obs_key, umap_path){"
            "umap_plot <- DimPlot(seurat, group.by = obs_key, "
            'reduction = "umap");'
            "ggsave(umap_path, umap_plot, width = 10, "
            "height = 6, dpi = 76)}"
        )
        r_umap = robjects.r(r_cmd)
        r_umap(seurat, obs_keys[0], umap_path)


def create_tsne_fig(seurat, atlas_path, atlas_info, args) -> None:
    """
    Display the TSNE of celltypes
    Search for the OBS variable which correspond to the celltype annotation
    :param path:
    :param adata:
    :param atlas_name:
    :param atlas_path:
    :return:
    """
    atlas_name = atlas_info[0]
    # Search if tsne reduction exists
    r = re.compile(".*tsne*.")
    r_names = robjects.r["names"]
    obsm_list = r_names(seurat)
    importr("ggplot2")
    if len(list(filter(r.match, obsm_list))) > 0:
        logger.debug(f"Create t-SNE figure for {atlas_name}")
        # Setting up figures directory
        tsne_path = os.path.join(
            folders.get_folder(args.path, folders.TSNE),
            atlas_name + checkatlas.TSNE_EXTENSION,
        )
        # Exporting tsne
        obs_keys = get_viable_obs_annot(seurat, args)
        r_cmd = (
            "tsne <- function(seurat, obs_key, tsne_path){"
            "tsne_plot <- DimPlot(seurat, group.by = obs_key, "
            'reduction = "tsne");'
            "ggsave(tsne_path, tsne_plot, width = 10, "
            "height = 6, dpi = 76)}"
        )
        r_tsne = robjects.r(r_cmd)
        r_tsne(seurat, obs_keys[0], tsne_path)


def metric_cluster(seurat, atlas_path, atlas_info, args) -> None:
    """
    Calc clustering metrics
    :param seurat:
    :param atlas_path:
    :param atlas_info:
    :param args:
    :return:
    """
    atlas_name = atlas_info[0]
    csv_path = os.path.join(
        folders.get_folder(args.path, folders.CLUSTER),
        atlas_name + checkatlas.METRIC_CLUSTER_EXTENSION,
    )
    header = ["Sample", "obs"] + args.metric_cluster
    df_cluster = pd.DataFrame(columns=header)
    obs_keys = get_viable_obs_annot(seurat, args)
    obsm_key_representation = "umap"
    if len(obs_keys) > 0:
        logger.debug(f"Calc clustering metrics for {atlas_name}")
        for obs_key in obs_keys:
            dict_line = {
                "Sample": [atlas_name + "_" + obs_key],
                "obs": [obs_key],
            }
            for metric in args.metric_cluster:
                logger.debug(
                    f"Calc {metric} for {atlas_name} "
                    f"with obs {obs_key} and obsm {obsm_key_representation}"
                )
                metric_value = metrics.calc_metric_cluster_seurat(
                    metric, seurat, obs_key, obsm_key_representation
                )
                dict_line[metric] = metric_value
            df_line = pd.DataFrame(dict_line)
            df_cluster = pd.concat(
                [df_cluster, df_line], ignore_index=True, axis=0
            )
        df_cluster.to_csv(csv_path, index=False, sep="\t")
    else:
        logger.debug(f"No viable obs_key was found for {atlas_name}")


def metric_annot(seurat, atlas_path, atlas_info, args) -> None:
    """
    Calc annotation metrics
    :param adata:
    :param atlas_path:
    :param atlas_info:
    :param args:
    :return:
    """
    atlas_name = atlas_info[0]
    csv_path = os.path.join(
        folders.get_folder(args.path, folders.ANNOTATION),
        atlas_name + checkatlas.METRIC_ANNOTATION_EXTENSION,
    )
    header = ["Sample", "Reference", "obs"] + args.metric_annot
    df_annot = pd.DataFrame(columns=header)
    obs_keys = get_viable_obs_annot(seurat, args)
    if len(obs_keys) > 1:
        logger.debug(f"Calc annotation metrics for {atlas_name}")
        if len(obs_keys) != 0:
            ref_obs = obs_keys[0]
            for i in range(1, len(obs_keys)):
                obs_key = obs_keys[i]
                dict_line = {
                    "Sample": [atlas_name + "_" + obs_key],
                    "Reference": [ref_obs],
                    "obs": [obs_key],
                }
                for metric in args.metric_annot:
                    logger.debug(
                        f"Calc {metric} for {atlas_name} "
                        f"with obs {obs_key} vs ref_obs {ref_obs}"
                    )
                    metric_value = metrics.calc_metric_annot_seurat(
                        metric, seurat, obs_key, ref_obs
                    )
                    dict_line[metric] = metric_value
                df_line = pd.DataFrame(dict_line)
                df_annot = pd.concat(
                    [df_annot, df_line], ignore_index=True, axis=0
                )
            df_annot.to_csv(csv_path, index=False, sep="\t")
    else:
        logger.debug(f"No viable obs_key was found for {atlas_name}")


def metric_dimred(seurat, atlas_path, atlas_info, args) -> None:
    """
    Calc dimensionality reduction metrics
    :param adata:
    :param atlas_path:
    :param atlas_info:
    :param args:
    :return:
    """
    atlas_name = atlas_info[0]
    csv_path = os.path.join(
        folders.get_folder(args.path, folders.DIMRED),
        atlas_name + checkatlas.METRIC_DIMRED_EXTENSION,
    )
    header = ["Sample", "obsm"] + args.metric_dimred
    df_dimred = pd.DataFrame(columns=header)
    # r_reduction = robjects.r(
    #     "reduc <- function(seurat, obsm_key){"
    #     " return(Embeddings(object = seurat, reduction = obsm_key))}"
    # )
    obsm_keys = get_viable_obsm(seurat, args)
    if len(obsm_keys) > 0:
        logger.debug(f"Calc dim red metrics for {atlas_name}")
        for obsm_key in obsm_keys:
            dict_line = {
                "Sample": [atlas_name + "_" + obsm_key],
                "obsm": [obsm_key],
            }
            for metric in args.metric_dimred:
                logger.debug(
                    f"Calc {metric} for {atlas_name} with obsm {obsm_key}"
                )
                # r_countmatrix = robjects.r(
                #     "mat <- function(seurat)
                #     { return(seurat@assays$RNA@counts)}"
                # )
                # high_dim_counts = ro.conversion.rpy2py(r_countmatrix(seurat))
                # low_dim_counts = ro.conversion.rpy2py(
                #    r_reduction(seurat, obsm_key)
                # )
                # metric_value = metrics.calc_metric_dimred(
                # metric, high_dim_counts, low_dim_counts)
                logger.warning(
                    "!!! Dim reduction metrics not available for Seurat"
                    " at the moment !!!"
                )
                # metric_value = -1
                # dict_line[metric] = str(metric_value)
            df_line = pd.DataFrame(dict_line)
            df_dimred = pd.concat(
                [df_dimred, df_line], ignore_index=True, axis=0
            )
        df_dimred.to_csv(csv_path, index=False, sep="\t")
    else:
        logger.debug(f"No viable obsm_key was found for {atlas_name}")



import csv
import inspect
import logging
import os
from datetime import datetime

from . import atlas, atlas_seurat, checkatlas_workflow, folders, multiqc

# try:
#     from . import atlas, folders, multiqc
# except ImportError:
#     import atlas
#     import folders
#     import multiqc


"""
checkatlas base module.
This is the principal module of the checkatlas project.

"""

EXTENSIONS = [".rds", ".h5ad", ".h5"]
CELLRANGER_FILE = "/outs/filtered_feature_bc_matrix.h5"
RSCRIPT = inspect.getfile(atlas).replace("atlas.py", "convertSeurat.R")
SUMMARY_EXTENSION = "_checkatlas_summ.tsv"
ADATA_EXTENSION = "_checkatlas_adata.tsv"
QC_FIG_EXTENSION = "_checkatlas_qc.png"
QC_EXTENSION = "_checkatlas_qc.tsv"
UMAP_EXTENSION = "_checkatlas_umap.png"
TSNE_EXTENSION = "_checkatlas_tsne.png"
METRIC_CLUSTER_EXTENSION = "_checkatlas_mcluster.tsv"
METRIC_ANNOTATION_EXTENSION = "_checkatlas_mannot.tsv"
METRIC_DIMRED_EXTENSION = "_checkatlas_mdimred.tsv"
METRIC_SPECIFICITY_EXTENSION = "_checkatlas_mspecificity.tsv"

logger = logging.getLogger("checkatlas")


def list_atlases(path: str) -> list:
    """
    List all atlases files in the path
    Detect .rds, .h5, .h5ad

    Args:
        path: Path for searching single-cell atlases.

    Returns:
        List of file atlases to check.
    """
    atlas_list = list()
    for root, dirs, files in os.walk(path):
        for file in files:
            for extension in EXTENSIONS:
                if file.endswith(extension):
                    atlas_list.append(os.path.join(root, file))
    return atlas_list


def get_atlas_name(atlas_path):
    """
    From atlas_path extract the atlas_name
    Args:
        atlas_path:
    Returns:

    """
    return os.path.splitext(os.path.basename(atlas_path))[0]


def get_atlas_extension(atlas_path):
    """
    From atlas_path extract the atlas file extension
    Args:
        atlas_path:
    Returns:

    """
    return os.path.splitext(os.path.basename(atlas_path))[1]


def clean_list_atlases(atlas_list, path) -> tuple:
    """
    Go through all files and detect Seurat or Scanpy Atlas
    Then:
    - Convert Seurat files to Scanpy
    - Clean Scanpy files
    Args:
        atlas_list:
    Returns:
         clean_list_atlas will only cleaned Scanpy atlas. A dict with
    these info ['Atlas_Name','Type','Atlas_file_extension',
    'Checkatlas_folder']
    """
    # Create dict with these info ['Atlas_Name','Type','Converted',
    # 'Checkatlas_path']
    clean_atlas_scanpy = dict()
    clean_atlas_seurat = dict()
    clean_atlas_cellranger = dict()
    for atlas_path in atlas_list:
        atlas_name = get_atlas_name(atlas_path)
        if atlas_path.endswith(".rds"):
            logger.debug(f"Include Atlas: {atlas_name} from {atlas_path}")
            info = [
                atlas_name,
                "Seurat",
                ".rds",
                os.path.dirname(atlas_path) + "/",
            ]
            clean_atlas_seurat[atlas_path] = info
        elif atlas_path.endswith(".h5"):
            # detect if its a cellranger output
            if atlas_path.endswith(CELLRANGER_FILE):
                atlas_h5 = atlas_path.replace(CELLRANGER_FILE, "")
                atlas_name = get_atlas_name(atlas_h5)
                logger.debug(f"Include Atlas: {atlas_name} from {atlas_path}")
                info = [
                    atlas_name,
                    "Cellranger",
                    ".h5",
                    os.path.dirname(atlas_h5) + "/",
                ]
                clean_atlas_cellranger[atlas_path] = info
        elif atlas_path.endswith(".h5ad"):
            logger.debug(f"Include Atlas: {atlas_name} from {atlas_path}")
            info = [
                atlas_name,
                "Scanpy",
                ".h5ad",
                os.path.dirname(atlas_path) + "/",
            ]
            clean_atlas_scanpy[atlas_path] = info
    # open file for writing, "w" is writing
    dict_file = open(
        os.path.join(folders.get_workingdir(path), "list_atlases.csv"), "w"
    )
    w = csv.writer(dict_file)
    # loop over dictionary keys and values
    for key, val in clean_atlas_scanpy.items():
        w.writerow([key, ",".join(val)])
    for key, val in clean_atlas_seurat.items():
        w.writerow([key, ",".join(val)])
    for key, val in clean_atlas_cellranger.items():
        w.writerow([key, ",".join(val)])
    dict_file.close()
    return clean_atlas_scanpy, clean_atlas_seurat, clean_atlas_cellranger


def get_pipeline_functions(module, args):
    """
    Using arguments of checkatlas program -> build
    the list of functions to run on each adata
    and seurat object
    Args:
        module:
        args:
    Returns:
         list of functions to run
    """
    checkatlas_functions = list()

    if not args.NOADATA:
        checkatlas_functions.append(module.create_anndata_table)
    if not args.NOQC:
        if "violin_plot" in args.qc_display:
            checkatlas_functions.append(module.create_qc_plots)
        if (
            "total-counts" in args.qc_display
            or "n_genes_by_counts" in args.qc_display
            or "pct_counts_mt" in args.qc_display
        ):
            checkatlas_functions.append(module.create_qc_tables)
    if not args.NOREDUCTION:
        checkatlas_functions.append(module.create_umap_fig)
        checkatlas_functions.append(module.create_tsne_fig)
    if not args.NOMETRIC:
        if len(args.metric_cluster) > 0:
            checkatlas_functions.append(module.metric_cluster)
        else:
            logger.debug(
                "No clustering metric was specified in --metric_cluster"
            )
        if len(args.metric_annot) > 0:
            checkatlas_functions.append(module.metric_annot)
        else:
            logger.debug(
                "No annotation metric was specified in --metric_annot"
            )
        if len(args.metric_dimred) > 0:
            checkatlas_functions.append(module.metric_dimred)
        else:
            logger.debug("No dim red metric was specified in --metric_dimred")
    # Create summary by default, it is ran at last so it marks
    # the end of the pipeline
    # This table is then used by the resume option
    checkatlas_functions.append(module.create_summary_table)
    return checkatlas_functions


def run(args):
    """
    Main function of checkatlas
    Run all functions for all atlases:
    - Clean files list by getting only Scanpy atlas (converted from Seurat
    if necessary)
    - Extract summary tables
    - Create UMAP and T-sne figures
    - Calculate every metrics

    Args:
        args:
    Returns:

    """
    logger.debug(f"Transform path to absolute:{args.path}")
    args.path = os.path.abspath(args.path)
    logger.debug(f"Check checkatlas folders in:{args.path}")
    folders.checkatlas_folders(args.path)

    logger.info("Searching Seurat, Cellranger and Scanpy files")
    atlas_list = list_atlases(args.path)
    # First clean atlas list and keep only the h5ad files
    (
        clean_atlas_scanpy,
        clean_atlas_seurat,
        clean_atlas_cellranger,
    ) = clean_list_atlases(atlas_list, args.path)
    logger.info(
        f"Found {len(clean_atlas_scanpy)} potential "
        f"scanpy files with .h5ad extension"
    )
    logger.info(
        f"Found {len(clean_atlas_seurat)} potential "
        f"seurat files with .rds extension"
    )
    logger.info(
        f"Found {len(clean_atlas_cellranger)} cellranger "
        f"file with .h5 extension"
    )

    # Put all atlases together in the list
    clean_atlas = dict(clean_atlas_scanpy)
    clean_atlas.update(clean_atlas_cellranger)
    clean_atlas.update(clean_atlas_seurat)

    if len(clean_atlas_cellranger) > 0:
        logger.debug("Install Seurat if needed")
        atlas_seurat.check_seurat_install()

    # Run all checkatlas analysis
    if args.nextflow == 0:
        logger.info(
            "--nextflow option not found: Run checkatlas workflow "
            "without Nextflow"
        )
        run_checkatlas(clean_atlas, args)
    else:
        clean_atlas.update(clean_atlas_seurat)
        logger.info(
            "--nextflow option found: Run checkatlas workflow with Nextflow"
        )
        logger.info(f"Use {args.nextflow} threads")
        checkatlas_workflow.create_checkatlas_worflows(clean_atlas, args)
        script_path = os.path.dirname(os.path.realpath(__file__))
        nextflow_main = os.path.join(script_path, "checkatlas_workflow.nf")
        yaml_files = os.path.join(
            folders.get_folder(args.path, folders.TEMP), "*.yaml"
        )

        # getting the current date and time
        current_datetime = datetime.now()
        current_time = current_datetime.strftime("%Y%d%m-%H%M%S")
        report_file = os.path.join(
            folders.get_workingdir(args.path),
            f"Nextflow_report-{current_time}.html",
        )
        timeline_file = os.path.join(
            folders.get_workingdir(args.path),
            f"Nextflow_timeline-{current_time}.html",
        )
        working_dir_nextflow = folders.get_folder(args.path, folders.NEXTFLOW)
        nextflow_cmd = (
            f"nextflow run -w "
            f"{working_dir_nextflow}"
            f" {nextflow_main} -queue-size {args.nextflow} --files "
            f'"{yaml_files}" -with-report {report_file}'
            f" -with-timeline {timeline_file}"
        )
        logger.debug(f"Execute: {nextflow_cmd}")
        script_path = os.path.dirname(os.path.realpath(__file__))
        nextflow_main = os.path.join(script_path, "checkatlas_workflow.nf")
        # Run Nextflow
        os.system(nextflow_cmd)
        logger.debug(f"Nextflow report saved in {report_file}")
        logger.debug(f"Nextflow timeline saved in {timeline_file}")

    if not args.NOMULTIQC:
        logger.info("Run MultiQC")
        multiqc.run_multiqc(args)


def run_checkatlas(clean_atlas, args):
    """
    Run Checkatlas pipeline for all Scanpy and Cellranger objects
    Args:
        clean_atlas_scanpy:
        args:
    Returns:

    """

    # List all functions to run
    pipeline_functions_scanpy = get_pipeline_functions(atlas, args)
    pipeline_functions_seurat = get_pipeline_functions(atlas_seurat, args)
    logger.debug(
        f"List of functions which will be ran "
        f"for each Seurat atlas: {pipeline_functions_scanpy}"
    )

    # Go through all atls
    for atlas_path, atlas_info in clean_atlas.items():
        atlas_name = atlas_info[0]
        # Load adata only if resume is not selected
        # and if csv_summary_path do not exist
        csv_summary_path = os.path.join(
            folders.get_folder(args.path, folders.SUMMARY),
            atlas_name + SUMMARY_EXTENSION,
        )
        if args.resume and os.path.exists(csv_summary_path):
            logger.debug(
                f"Skip {atlas_name} summary file already "
                f"exists: {csv_summary_path}"
            )
        else:
            if atlas_info[1] == "Seurat":
                seurat = atlas_seurat.read_atlas(atlas_path, atlas_info)
                logger.info(
                    f"Run checkatlas pipeline for {atlas_name} Seurat atlas"
                )
                # Run pipeline functions
                for function in pipeline_functions_seurat:
                    function(seurat, atlas_path, atlas_info, args)
            else:
                adata = atlas.read_atlas(atlas_path, atlas_info)
                # Clean adata
                adata = atlas.clean_scanpy_atlas(adata, atlas_info)
                logger.info(
                    f"Run checkatlas pipeline for {atlas_name} Scanpy atlas"
                )
                # Run pipeline functions
                for function in pipeline_functions_scanpy:
                    function(adata, atlas_path, atlas_info, args)


if __name__ == "__main__":
    path = "/Users/christophebecavin/Documents/testatlas/"
    # atlas_path = "/Users/christophebecavin/Documents/testatlas/"
    atlas_info = ["test_version", "Scanpy", ".h5ad", "data/test_version.h5ad"]
    # folders.checkatlas_folders(path)
    # atlas_list = list_atlases(path)
    # clean_atlas_dict = clean_list_atlases(atlas_list, path)
    #
    # for atlas_path, atlas_info in clean_atlas_dict.items():
    #     print(atlas_path, atlas_info)
    #     adata = read_atlas(atlas_path, atlas_info)
    #     print(adata is not None)
    #     if adata is not None:
    #         adata = atlas.clean_scanpy_atlas(adata, atlas_info)
    #         atlas.create_summary_table(adata, atlas_path, atlas_info, path)
    #         atlas.create_anndata_table(adata, atlas_path, atlas_info, path)
    #         # atlas.create_qc_plots(adata, atlas_path, atlas_info, path)
    #         atlas.create_umap_fig(adata, atlas_path, atlas_info, path)
    #         # atlas.create_tsne_fig(adata, atlas_path, atlas_info, path)
    #         atlas.metric_cluster(adata, atlas_path, atlas_info, path)
    #         atlas.metric_annot(adata, atlas_path, atlas_info, path)
    #         atlas.metric_dimred(adata, atlas_path, atlas_info, path)
# atlas_path = '/Users/christophebecavin/Documents/testatlas/hca/
# HCA_Barbry_Grch38_Raw_filter_Norm.h5ad'
# atlas_path = '/Users/christophebecavin/Documents/testatlas/Endothelial.h5ad'
# atlas_name = get_atlas_name(atlas_path)
# print('Read atlas')
#     adata = read_atlas(atlas_path, atlas_info)
#     sc.pp.subsample(adata, fraction=0.05)
#     adata.write('/Users/christophebecavin/Documents/testatlas/Endothelial_lite.h5ad')
#     print(adata)
# print('Calc QC')
# figure_path = '/Users/christophebecavin/Documents/testatlas/'
# atlas.create_qc_plots(adata, atlas_path, atlas_name, figure_path)



import argparse  # pragma: no cover
import logging
import os.path

import yaml

from . import checkatlas  # pragma: no cover
from . import atlas, atlas_seurat, folders

logger = logging.getLogger("checkatlas-workflow")
FUNCTION_TYPE = [
    "summary",
    "qc",
    "metric_cluster",
    "metric_annot",
    "metric_dimred",
]


def workflow():
    # Set up logging
    logging.basicConfig(format="|--- %(levelname)-8s %(message)s")

    parser = argparse.ArgumentParser(
        prog="checkatlas-workflow",
        usage="checkatlas-workflow summary my_workflow.yaml",
        description="CheckAtlas-workflow is the script ran by checkatlas "
        "for each detected atlas in Scanpy, Seurat, "
        "and CellRanger format.",
        epilog="!!! You are not supposed to run checkatlas-workflow "
        "by yourself. It is the main program (checkatlas) who"
        " ran it using Nextflow !!!",
    )

    # All Program arguments
    # main_options = parser.add_argument_group("Main arguments")
    parser.add_argument(
        "type",
        type=str,
        help="Provide the type of function to run in the workflow."
        "Different type of function : summary, qc, metric_cluster, "
        "metric_annot, metric_dimred",
        default="",
    )

    parser.add_argument(
        "workflow",
        type=str,
        help="Provide a workflow file with all checkatlas-workflow "
        "arguments. The workflow config file are created by "
        "checkatlas directly and saved in temp/ folder.",
        default="",
    )
    parser.add_argument(
        "--atlas_name",
        type=str,
        help="Name of the atlas.",
        default=".",
    )
    parser.add_argument(
        "--atlas_type",
        type=str,
        help="Required argument: The type of atlas "
        "Scanpy, CellRanger and Seurat."
        "Values : scanpy, seurat, cellranger",
        default="scanpy",
    )
    parser.add_argument(
        "--atlas_path",
        type=str,
        help="Path of the atlas.",
        default=".",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Print out all debug messages.",
    )

    # Pipeline arguments
    pipeline_options = parser.add_argument_group("Manage checkatlas pipeline")
    pipeline_options.add_argument(
        "-na",
        "--NOADATA",
        action="store_true",
        help="Do not Create adata summary tables.",
    )
    pipeline_options.add_argument(
        "-nq",
        "--NOQC",
        action="store_true",
        help="Do not produce any quality control figures or tables.",
    )
    pipeline_options.add_argument(
        "-nr",
        "--NOREDUCTION",
        action="store_true",
        help="Do not Produce UMAP and t-SNE figures.",
    )
    pipeline_options.add_argument(
        "-nm",
        "--NOMETRIC",
        action="store_true",
        help="Do not calculate any metric.",
    )
    pipeline_options.add_argument(
        "-nc", "--NOMULTIQC", action="store_true", help="Do not run multiqc."
    )

    # Arguments linked to QC
    qc_options = parser.add_argument_group("QC options")
    qc_options.add_argument(
        "--qc_display",
        nargs="+",
        type=str,
        default=[
            "violin_plot",
            "total_counts",
            "n_genes_by_counts",
            "pct_counts_mt",
        ],
        help="List of QC to display. "
        "Available qc = violin_plot, total_counts, "
        "n_genes_by_counts, pct_counts_mt. "
        "Default: --qc_display violin_plot total_counts "
        "n_genes_by_counts pct_counts_mt",
    )

    # Arguments linked to metric
    metric_options = parser.add_argument_group("Metric options")
    metric_options.add_argument(
        "--obs_cluster",
        nargs="+",
        type=str,
        default=atlas.OBS_CLUSTERS,
        help="List of obs from the adata file to "
        "use in the clustering metric calculus."
        "Example: --obs_cluster celltype leuven seurat_clusters",
    )
    metric_options.add_argument(
        "--metric_cluster",
        nargs="+",
        type=str,
        default=["silhouette", "davies_bouldin"],
        help="List of clustering metrics to calculate."
        "To get a complete list of metrics, look here:"
        "https://github.com/becavin-lab/checkatlas/"
        "blob/main/checkatlas/metrics/metrics.py"
        " Example: --metric_cluster silhouette davies_bouldin",
    )
    metric_options.add_argument(
        "--metric_annot",
        nargs="+",
        type=str,
        default=["rand_index"],
        help="List of clustering metrics to calculate."
        "To get a complete list of metrics, look here:"
        "https://github.com/becavin-lab/checkatlas/blob/"
        "main/checkatlas/metrics/metrics.py"
        " Example: --metric_annot rand_index",
    )
    metric_options.add_argument(
        "--metric_dimred",
        nargs="+",
        type=str,
        default=["kruskal_stress"],
        help="List of dimensionality reduction metrics to calculate."
        "To get a complete list of metrics, look here:"
        "https://github.com/becavin-lab/checkatlas/blob/"
        "main/checkatlas/metrics/metrics.py"
        " Example: --metric_dimred kruskal_stress",
    )

    # Parse all args
    args = parser.parse_args()

    # If a config file was provided, load the new args !
    # Normally a config file is always provided
    if args.workflow != "":
        logger.info(
            "Read workflow config file {} to get new checkatlas "
            "configs".format(args.workflow)
        )
        args = load_arguments(args, args.workflow)

    # Set logger level
    if args.debug:
        logger.setLevel(getattr(logging, "DEBUG"))
    else:
        logger.setLevel(getattr(logging, "INFO"))

    logger.debug(f"Program arguments: {args}")

    # !!! atlas_info should be removed in the future !
    atlas_info = [args.atlas_name, args.atlas_type, ".rds", args.atlas_path]
    csv_summary_path = os.path.join(
        folders.get_folder(args.path, folders.SUMMARY),
        args.atlas_name + checkatlas.SUMMARY_EXTENSION,
    )
    if args.resume and os.path.exists(csv_summary_path):
        logger.debug(
            f"Skip {args.atlas_name} summary file already "
            f"exists: {csv_summary_path}"
        )
    else:
        if args.atlas_type == "Seurat":
            pipeline_functions = get_workflow_functions(atlas_seurat, args)
            seurat = atlas_seurat.read_atlas(args.atlas_path, atlas_info)
            # Run pipeline functions
            for function in pipeline_functions:
                function(seurat, args.atlas_path, atlas_info, args)
        else:
            pipeline_functions = get_workflow_functions(atlas, args)
            adata = atlas.read_atlas(args.atlas_path, atlas_info)
            # Clean adata
            adata = atlas.clean_scanpy_atlas(adata, atlas_info)
            # Run pipeline functions
            for function in pipeline_functions:
                function(adata, args.atlas_path, atlas_info, args)


def get_workflow_functions(module, args):
    """
    Using arguments of checkatlas-workflow program -> build
    the list of functions to run on each adata
    and seurat object
    :param args:
    :return: list of functions to run
    """
    checkatlas_functions = list()
    if args.type == "summary":
        if not args.NOADATA:
            checkatlas_functions.append(module.create_anndata_table)
        if not args.NOREDUCTION:
            checkatlas_functions.append(module.create_umap_fig)
            checkatlas_functions.append(module.create_tsne_fig)
        checkatlas_functions.append(module.create_summary_table)
    elif args.type == "qc" and not args.NOQC:
        if "violin_plot" in args.qc_display:
            checkatlas_functions.append(module.create_qc_plots)
        if (
            "total-counts" in args.qc_display
            or "n_genes_by_counts" in args.qc_display
            or "pct_counts_mt" in args.qc_display
        ):
            checkatlas_functions.append(module.create_qc_tables)
    elif args.type == "metric_cluster" and not args.NOMETRIC:
        if len(args.metric_cluster) > 0:
            checkatlas_functions.append(module.metric_cluster)
    elif args.type == "metric_annot" and not args.NOMETRIC:
        if len(args.metric_annot) > 0:
            checkatlas_functions.append(module.metric_annot)
    elif args.type == "metric_annot" and not args.NOMETRIC:
        if len(args.metric_dimred) > 0:
            checkatlas_functions.append(module.metric_dimred)
    return checkatlas_functions


def create_checkatlas_worflows(clean_atlas, args):
    """
    Create Checkatlas workflow config files for each atlas
    :param clean_atlas:
    :param args:
    :return:
    """
    # Create workflow config files
    temp_path = folders.get_folder(args.path, folders.TEMP)
    for atlas_path, atlas_info in clean_atlas.items():
        atlas_name = atlas_info[0]
        atlas_dict = dict()
        atlas_dict["atlas_name"] = atlas_name
        atlas_dict["atlas_path"] = atlas_path
        atlas_dict["atlas_type"] = atlas_info[1]

        config_path = os.path.join(
            temp_path, f"Checkatlas_workflow_{atlas_name}.yaml"
        )
        with open(config_path, "w") as config_file:
            yaml.dump(atlas_dict, config_file)
            yaml.dump(args.__dict__, config_file)
        logger.info(f"Workflow config file saved in : {config_path}")


def load_arguments(args, yaml_name):
    """
    Load all args from a yaml file.

    :param args:
    :param yaml_name:
    :return: args
    """
    with open(yaml_name, "r") as config_file:
        yaml_args = yaml.load(config_file, Loader=yaml.FullLoader)
        arg_dict = args.__dict__
        for key, value in yaml_args.items():
            if isinstance(value, list):
                arg_dict[key] = list()
                for v in value:
                    arg_dict[key].append(v)
            else:
                arg_dict[key] = value
        return args


if __name__ == "__main__":  # pragma: no cover
    workflow()



from sklearn.metrics import adjusted_rand_score


def run(annotation, ref_annotation):
    """



    :param annotation:
    :param ref_annotation:
    :return:
    """
    return adjusted_rand_score(annotation, ref_annotation)



import logging
import os

working_dir = "checkatlas_files"
SUMMARY = "summary"
ANNDATA = "adata"
QC = "qc"
QC_FIG = "violin"
UMAP = "umap"
TSNE = "tsne"
CLUSTER = "cluster"
ANNOTATION = "annotation"
DIMRED = "dimred"
SPECI = "specificity"
TEMP = "temp"
NEXTFLOW = "temp/nextflow"

dict_folder = {
    SUMMARY: SUMMARY,
    ANNDATA: ANNDATA,
    QC: QC,
    QC_FIG: QC_FIG,
    UMAP: UMAP,
    TSNE: TSNE,
    CLUSTER: CLUSTER,
    ANNOTATION: ANNOTATION,
    DIMRED: DIMRED,
    SPECI: SPECI,
    TEMP: TEMP,
    NEXTFLOW: NEXTFLOW,
}

logger = logging.getLogger("checkatlas")


def get_workingdir(path):
    return os.path.join(path, working_dir)


def get_folder(path, key_folder):
    return os.path.join(get_workingdir(path), dict_folder[key_folder])


def checkatlas_folders(path):
    """
    Check in path if the different checkatlas folders exists.<br>
    Create them if needed.
    :param path:
    :return: None
    """
    global_path = get_workingdir(path)
    if not os.path.exists(global_path):
        os.mkdir(global_path)

    for key_folder in dict_folder.keys():
        temp_path = os.path.join(global_path, key_folder)
        if not os.path.exists(temp_path):
            logger.debug(f"Create folder: {temp_path}")
            os.mkdir(temp_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['checkatlas',
 'checkatlas.metrics',
 'checkatlas.metrics.annot',
 'checkatlas.metrics.cluster',
 'checkatlas.metrics.dimred',
 'checkatlas.metrics.specificity']

package_data = \
{'': ['*']}

install_requires = \
['llvmlite>=0.39.1,<0.40.0',
 'matplotlib>=3.7.0,<4.0.0',
 'multiqc>=1.14,<2.0',
 'nextflow>=22.10.7,<23.0.0',
 'numba>=0.56.4,<0.57.0',
 'numpy>=1.23.5,<2.0.0',
 'rpy2>=3.5.8,<4.0.0',
 'scanpy>=1.9.1,<2.0.0',
 'scikit-learn>=1.2.1,<2.0.0',
 'types-pyyaml>=6.0.12.6,<7.0.0.0']

entry_points = \
{'console_scripts': ['checkatlas = checkatlas.__main__:main',
                     'checkatlas-workflow = '
                     'checkatlas.checkatlas_workflow:workflow']}

setup_kwargs = {
    'name': 'checkatlas',
    'version': '0.2.2',
    'description': 'One liner tool to check the quality of your single-cell atlases.',
    'long_description': "\n# CheckAtlas\n\n[![codecov](https://codecov.io/gh/becavin-lab/checkatlas/branch/main/graph/badge.svg?token=checkatlas_token_here)](https://codecov.io/gh/becavin-lab/checkatlas)\n[![CI](https://github.com/becavin-lab/checkatlas/actions/workflows/tests.yml/badge.svg)](https://github.com/becavin-lab/checkatlas/actions/workflows/tests.yml)\n[![Documentation Status](https://readthedocs.org/projects/checkatlas/badge/?version=latest)](https://checkatlas.readthedocs.io/en/latest/?badge=latest)\n\nCheckAtlas is a one liner tool to check the quality of your single-cell atlases. For every atlas, it produces the\nquality control tables and figures which can be then processed by multiqc. CheckAtlas is able to load Scanpy, Seurat,\nand CellRanger files.\n\n\n## Summary\n\n### Parse Scanpy, Seurat and CellRanger objects\n\nThe checkatlas pipeline start with a fast crawl through your working directory. It detects Seurat (.rds), Scanpy (.h5ad) or cellranger (.h5) atlas files.\n\n\n### Create checkatlas summary files\n\nGo through all atlas files and produce summary information:\n\n- All basic QC (nRNA, nFeature, ratio_mito)\n- General information (nbcells, nbgenes, nblayers)\n- All elements in atlas files (obs, obsm, uns, var, varm)\n- Reductions (pca, umap, tsne)\n- All metrics (clustering, annotation, dimreduction, specificity)\n\n### Parse checkatlas files in MultiQC\n\n   Update MultiQC project to add checkatlas parsing. Dev project in: https://github.com/becavin-lab/MultiQC/tree/checkatlas\n\nhttps://checkatlas.readthedocs.io/en/latest/\n\n## Examples\n\n1. Evaluate and compare different atlases: https://github.com/becavin-lab/checkatlas/blob/3a4f88e94716c09a3b9c86010f570743a5855461/examples/Atlas_comparison.ipynb\n\nhttps://checkatlas.readthedocs.io/en/stable/CheckAtlas_example_1/CheckAtlas_example_1.html\n\n2. Evaluate different version of your atlas: https://github.com/becavin-lab/checkatlas/blob/3a4f88e94716c09a3b9c86010f570743a5855461/examples/Version_comparison.ipynb\n\nhttps://checkatlas.readthedocs.io/en/stable/CheckAtlas_example_2/CheckAtlas_example_2.html\n\n3. Explore Scanpy, Seurat and CellRanger objects in your folder: https://github.com/becavin-lab/checkatlas/blob/main/examples/AtlasType_comparison.ipynb\n\nhttps://checkatlas.readthedocs.io/en/stable/CheckAtlas_example_3/CheckAtlas_example_3.html\n\n## Installation\n\nCheckAtlas can be downloaded from PyPI. However, the project is in an early development phase. We strongly recommend to use the developmental version.\n\n### Install checkatlas development version\n\n```bash\ngit clone git@github.com:becavin-lab/checkatlas.git\npip install checkatlas/.\n```\n\nInstall MultiQC with checkatlas file management. This version of MultiQC is available at checkatlas branch of github.com:becavin-lab/MultiQC.\n\n```bash\ngit clone git@github.com:becavin-lab/MultiQC.git\ncd MultiQC/\ngit checkout checkatlas\npip install .\n```\n\n### Install it from PyPI\n\n```bash\npip install checkatlas\n```\n\n### Install Seurat\n\nTo be able to manage seurat file, rpy2 should have Seurat installed. The easiest way is to put all checkatlas requirements in a conda environment and add r-seurat.\n\n```bash\nconda create -n checkatlas python=3.9\npip install checkatlas\nconda install -c bioconda r-seurat\n```\n\nOr, open R in checkatlas environment (the one where you ran 'pip install') and install Seurat.\n\n```bash\n% R\n> install.packages('Seurat')\n> library(Seurat)\n```\n\n\n## Usage\n\nThe one liner way to run checkatlas is the following: \n\n```bash\n$ cd your_search_folder/\n$ python -m checkatlas .\n#or\n$ checkatlas .\n```\n\nOr run it inside your python workflow.\n\n```py\nfrom checkatlas import checkatlas\ncheckatlas.run(path, atlas_list, multithread, n_cpus)\n```\n\n\n## Development\n\nRead the [CONTRIBUTING.md](docs/contributing.md) file.\n\nProject developed thanks to the project template : (https://github.com/rochacbruno/python-project-template/)\n\n",
    'author': 'becavin-lab',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://checkatlas.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Created on Thu Jun  4 16:38:47 2020

@author: antoinecollin
"""

from typing import Collection, Iterable, Optional

import anndata
import get_data
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from compute import one_v_max_matrix
from scanpy.plotting._tools.scatterplots import _panel_grid

ROOTDIR = r"C:\Users\ipmc\Documents\Package_Metrics\scanpy\scanpy"
FILENAME = "HCA_Barbry_Hg19_seurat.h5ad"
filename = "\\".join([ROOTDIR, "datasets", FILENAME])
annick = anndata.read_h5ad(filename)


def gene_expr_distribution(
    adata: anndata,
    gene: str,
    ax,
    partition_key: str = "CellType",
    celltype: Optional[str] = None,
):
    """
    Plot the distribution of the average gene expression across celltypes

    Parameters
    ----------
    adata
        The corrected expression data.

    gene
        Gene to plot

    ax
        axe on which to plot

    partition_key
        The key in adata.obs corresponding to the annotations to be used.

    celltype
        Celltype to highlight in the rugplot.

    Returns
    -------

    """
    average_by_celltype = get_data.get_average_celltype_counts(
        adata=adata, partition_key=partition_key
    )
    out = sns.distplot(
        average_by_celltype[gene],
        kde=False,
        rug=True,
        ax=ax,
        hist_kws={
            "histtype": "stepfilled",
            "facecolor": "royalblue",
            "edgecolor": "black",
            "alpha": 0.8,
            "linewidth": 1,
        },
    )
    indent = ""
    for spe in ["shannon", "tau", "gini"]:
        spe_to_plot = round(
            get_data.get_spe(adata, spe_metric=spe, partition_key="CellType")[
                gene
            ],
            2,
        )
        out = plt.annotate(
            spe + " = " + str(spe_to_plot) + indent,
            xy=(0.79, 0.70),
            xycoords="axes fraction",
        )
        indent += "\n"
    if isinstance(celltype, str):
        out = sns.rugplot(
            [average_by_celltype.loc[celltype, gene]], color="red"
        )
        out = plt.annotate(
            celltype,
            xy=[average_by_celltype.loc[celltype, gene], 0.06],
            xycoords=("data", "axes fraction"),
            annotation_clip=False,
        )
    ax.set(xlabel=None, ylabel="Number of celltypes")
    ax.set_title(gene, fontsize=10)
    return out


def marker_genes_distribution(
    adata: anndata,
    gene_list: list,
    celltype: Optional[str],
    partition_key: str = "CellType",
    style="seaborn-white",
):
    """
    Display the distribution of the expressions of the genes across all
    celltypes
    Parameters
    ----------
    adata
        The corrected expression data.

    gene_list
        List of genes to plot, usually marker genes

    partition_key
        The key in adata.obs corresponding to the annotations to be used.

    celltype
        Celltype to highlight in the rugplot

    Returns
    -------

    """
    plt.style.use(style)
    n_panels = len(gene_list)
    f, grid = _panel_grid(
        hspace=0.25, wspace=0.25, ncols=4, num_panels=n_panels
    )
    for i in range(n_panels):
        gene = gene_list[i]
        ax = f.add_subplot(grid[i])
        gene_expr_distribution(
            adata=adata,
            gene=gene,
            ax=ax,
            partition_key=partition_key,
            celltype=celltype,
        )


def distrib_one_v_max(
    adata: anndata,
    celltype: str,
    ax,
    gene_highlight: Iterable[str],
    partition_key: str = "CellType",
):
    """

    Parameters
    ----------
    adata
        The corrected expression data.

    celltype
        Celltype to be plotted

    gene_highlight
        List of genes to highlight on the plot, usually marker genes

    partition_key
        The key in adata.obs corresponding to the annotations to be used.


    Returns
    -------

    """
    one_v_max = one_v_max_matrix(adata, partition_key=partition_key)
    to_plot = one_v_max.loc[celltype]
    to_plot_trunc = to_plot[
        to_plot > 1
    ]  # The only relevant one_v_max are the ones >1
    gene_highlight = list(set(gene_highlight) & set(list(to_plot_trunc.index)))
    if gene_highlight:
        to_plot_highlight = to_plot_trunc[gene_highlight]
        to_plot_trunc = to_plot_trunc.drop(gene_highlight)
        out = sns.stripplot(
            y=to_plot_trunc, ax=ax, orient="v", s=3, linewidth=0.25
        )
        out = sns.stripplot(
            y=to_plot_highlight, ax=ax, orient="v", color="red", linewidth=0.5
        )
        out.axes.set_xticks([])
        if len(gene_highlight) == 1:
            spe_ovm = round(to_plot_highlight[0], 2)
            out.axes.set_title(
                label=f"{gene_highlight[0]} \n "
                f"{to_plot_trunc.name} onevmax={spe_ovm}",
                loc="center",
                fontsize="small",
            )
        else:
            out.axes.set_title(
                label=f"{to_plot_trunc.name}", loc="center", fontsize="small"
            )
        out.axes.set_ylabel(ylabel="")
    else:
        out = sns.stripplot(
            y=to_plot_trunc, ax=ax, orient="v", s=3, linewidth=0.25
        )
        out.axes.set_title(
            label=f"{to_plot_trunc.name}", loc="center", fontsize="small"
        )
        out.axes.set_ylabel(ylabel="")
    return out


# Not used
def one_v_max_celltypes(
    adata: anndata,
    celltype_list: Optional[Collection[str]] = None,
    partition_key: str = "CellType",
):
    """
    Plots the one_v_max stripplot across the specified celltypes
    in the same plot


    Parameters
    ----------
    adata
        The corrected expression data.

    celltype_list
        celltypes to plot. Default plots every celltypes

    partition_key
        The key in adata.obs corresponding to the annotations to be used.

    Returns
    -------

    """
    one_v_max = one_v_max_matrix(adata, partition_key=partition_key)
    one_v_max["celltype"] = one_v_max.index
    mean_melt = pd.melt(
        one_v_max,
        id_vars="celltype",
        value_vars=list(one_v_max.columns[:-1]),
        var_name="genes",
        value_name="expression",
    )
    mean_melt = mean_melt[mean_melt["expression"] > 1]
    out = sns.stripplot(
        x=mean_melt["celltype"], y=mean_melt["expression"], s=2
    )
    out.set(yscale="log")
    return out


# Used
def one_v_max_genelist(
    adata: anndata, gene_list: list, partition_key: str = "CellType"
):
    """
    For each gene, plots the one_v_max stripplot in the celltype where it is
    maximum.

    Parameters
    ----------
    adata
    gene_list
    partition_key

    Returns
    -------

    """
    one_v_max = one_v_max_matrix(adata, partition_key=partition_key)
    n_panels = len(gene_list)
    f, grid = _panel_grid(
        hspace=0.25, wspace=0.25, ncols=4, num_panels=n_panels
    )
    for i in range(n_panels):
        gene = gene_list[i]
        ax = f.add_subplot(grid[i])
        ax.set(yscale="log")
        ax.set_yticks = [5 ** (n % 2) * 10 ** (n // 2) for n in range(10)]
        max_celltype = pd.to_numeric(one_v_max[gene]).idxmax()
        distrib_one_v_max(
            adata=adata,
            celltype=max_celltype,
            ax=ax,
            gene_highlight=[gene],
            partition_key=partition_key,
        )


# Not used
def one_v_max_celltypes_sep(
    adata: anndata,
    celltype_list: list,
    gene_highlight=None,
    partition_key: str = "CellType",
):
    """
    Plots the one_v_max distribution across the specified celltypes in
    separate plots

    Parameters
    ----------
    adata
        The corrected expression data.

    celltype_list
        celltypes to plot

    partition_key
        The key in adata.obs corresponding to the annotations to be used.


    Returns
    -------

    """
    n_panels = len(celltype_list)
    f, grid = _panel_grid(
        hspace=0.25, wspace=0.25, ncols=4, num_panels=n_panels
    )
    for i in range(n_panels):
        celltype = celltype_list[i]
        ax = f.add_subplot(grid[i])
        ax.set(yscale="log")
        distrib_one_v_max(
            adata=adata,
            celltype=celltype,
            ax=ax,
            gene_highlight=gene_highlight,
            partition_key=partition_key,
        )


def spec_distrib(
    adata: anndata,
    spe_name: str,
    ax,
    partition_key: str = "CellType",
):
    """
    Plots the distribution of the specified gene specificity

    Parameters
    ----------
    adata


    spe_name
        Specificity to plot

    ax

    partition_key


    Returns
    -------

    """
    dist_to_plot = get_data.get_spe(
        adata=adata, spe_metric=spe_name, partition_key=partition_key
    )
    out = sns.distplot(dist_to_plot, ax=ax)
    return out


def all_spec_distrib(adata: anndata, partition_key: str = "CellType"):
    f, (ax1, ax2, ax3) = plt.subplots(3, 1)
    for ax, spe in [(ax1, "shannon"), (ax2, "tau"), (ax3, "gini")]:
        spec_distrib(
            adata=adata, spe_name=spe, ax=ax, partition_key=partition_key
        )


# TODO
def marked_celltype_UMAP(
    adata: anndata, gene: str, marked_celltypes: Collection[str]
):
    """

    Parameters
    ----------
    adata
    gene
    marked_celltypes

    Returns
    -------

    """


# TODO
def gene_distrib_celltype(adata, gene, celltype):
    """

    Parameters
    ----------
    adata
    gene
    celltype

    Returns
    -------

    """


# if __name__ == "__main__" :
#     a = ['SERPINB4','SERPINB3','LY6D','CLCA2','MT1X','CCND1','AKR1C3',
#     'AQP3','DAPL1','NUPR1','S100A2','KRT5','CSRP2','IGFBP3','CALML3',
#     'TSPAN1','GPX2','FAM3B','EPAS1','TXN','SERPINB13','SUSD4','FAM84A']
#
#     one_v_max_celltypes(barbry,celltype_list=cl,partition_key='CellType')
#
#     ovm = one_v_max_matrix(barbry)
#
#     tp = ovm.loc['Basal']
#     tpt =tp[tp>1]
#
#     one_v_max_genelist(barbry, a, partition_key='CellType')
#
#     all_spec_distrib(barbry, partition_key='CellType')
